<?php

namespace Drupal\content_moderation_bulk_state_change\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\content_moderation\ModerationInformationInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Url;
use Drupal\node\NodeInterface;
use Drupal\user\UserStorageInterface;
use Drupal\workflows\TransitionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides a form for moving multiple entity page workflow states.
 *
 * @internal
 */
final class BulkChangeWorkflowStageForm extends ConfirmFormBase {

  /**
   * The temp store factory.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStoreFactory
   */
  protected $tempStoreFactory;

  /**
   * The user storage.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected $userStorage;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Moderation Information service.
   *
   * @var \Drupal\content_moderation\ModerationInformationInterface
   */
  protected ModerationInformationInterface $moderationInformation;

  /**
   * The request stack service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * Constructs a new BulkChangeWorkflowStageForm.
   *
   * @param \Drupal\Core\TempStore\PrivateTempStoreFactory $temp_store_factory
   *   The temp store factory.
   * @param \Drupal\user\UserStorageInterface $user_storage
   *   The user storage.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\content_moderation\ModerationInformationInterface $moderation_information
   *   The moderation information.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   */
  public function __construct(PrivateTempStoreFactory $temp_store_factory, UserStorageInterface $user_storage, EntityTypeManagerInterface $entity_type_manager, ModerationInformationInterface $moderation_information, RequestStack $request_stack, TimeInterface $time, ConfigFactoryInterface $config_factory) {
    $this->tempStoreFactory = $temp_store_factory;
    $this->userStorage = $user_storage;
    $this->entityTypeManager = $entity_type_manager;
    $this->moderationInformation = $moderation_information;
    $this->requestStack = $request_stack;
    $this->time = $time;
    $this->configFactory = $config_factory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('tempstore.private'),
      $container->get('entity_type.manager')->getStorage('user'),
      $container->get('entity_type.manager'),
      $container->get('content_moderation.moderation_information'),
      $container->get('request_stack'),
      $container->get('datetime.time'),
      $container->get('config.factory'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'bulk_change_workflow_stage_form';
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Change workflow stage');
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl(): ?Url {
    $destination = $this->requestStack->getCurrentRequest()->get('destination');
    if ($destination) {
      $url = Url::fromUserInput('/' . ltrim($destination, '/'));
      return Url::fromRoute('view.content.page_1');
    }
    // If no destination param, don't show cancel button.
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText(): string {
    return $this->t('Confirm');
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription(): string {
    return 'Click confirm to change the workflow stage of your entities.';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $entities = $this->tempStoreFactory
      ->get('bulk_change_workflow_stage')
      ->get($this->currentUser()->id());

    $errors = [];

    if (!$entities) {
      $errors[] = $this->t('No entities found');
    }

    // Redirect user if they do not have permission to edit entities.
    if (!$this->userHasEditPermissions($entities)) {
      $errors[] = $this->t('You do not have permission to edit all the selected entities.');
    }

    // Redirect user if entities are not all the same workflow.
    if (!$this->entitiesAllSameWorkflow($entities)) {
      $errors[] = $this->t('All entities need to use the same workflow');
    }

    // Redirect user if entities are not all the same workflow stage.
    if (!$this->entitiesAllSameWorkflowStage($entities)) {
      $errors[] = $this->t('All entities need to be in the same workflow stage');
    }

    // Redirect if any errors in user selection.
    if ($errors) {
      foreach ($errors as $error) {
        $this->messenger()->addMessage($error, MessengerInterface::TYPE_ERROR);
      }
      return $this->redirectToListing();
    }

    // Get first entity from array and workflow details.
    $entity = reset($entities);
    $latest_revision_id = $this->entityTypeManager->getStorage('node')->getLatestRevisionid($entity->id());
    $entity = $this->entityTypeManager->getStorage('node')->loadRevision($latest_revision_id);
    $current_stage = $entity->get('moderation_state')->value;
    $workflow_states = $this->moderationInformation->getWorkflowForEntity($entity)->getTypePlugin()->getStates();
    $current_stage_label = $workflow_states[$current_stage]->label();
    $available_to_states = $this->getAvailableToStates($entity);

    $entity_titles = [];
    $form['entities'] = ['#tree' => TRUE];
    foreach ($entities as $entity) {
      $nid = $entity->id();
      $entity_titles[$nid] = $entity->label();

      $form['entities'][$nid] = [
        '#type' => 'hidden',
        '#value' => $nid,
      ];
    }

    $form['account']['entities'] = [
      '#theme' => 'item_list',
      '#items' => $entity_titles,
      '#prefix' => $this->t('<p><strong>Entities:</strong></p>'),
      '#suffix' => $this->t('<hr/>'),
    ];

    $form['current_details'] = [
      '#type' => 'markup',
      '#markup' => $this->t('<p><strong>Current stage: </strong>@current_stage</p>', ['@current_stage' => $current_stage_label]),
    ];

    if ($available_to_states) {
      $form['transition_to'] = [
        '#type' => 'select',
        '#title' => $this->t('Move entity to:'),
        '#options' => $available_to_states,
        '#description' => $this->t('Please select an option from the dropdown.'),
        '#required' => TRUE,
      ];
    }

    $form['operation'] = ['#type' => 'hidden', '#value' => 'cancel'];
    $form = parent::buildForm($form, $form_state);

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $current_user_id = $this->currentUser()->id();

    $nodes = $this->tempStoreFactory->get('bulk_change_workflow_stage')->get($this->currentUser()->id());

    // Clear out the accounts from the temp store.
    $this->tempStoreFactory->get('bulk_change_workflow_stage')->delete($current_user_id);

    /** @var \Drupal\node\NodeStorageInterface $node_storage */
    $node_storage = $this->entityTypeManager->getStorage('node');

    if ($form_state->getValue('confirm')) {
      foreach ($nodes as $node) {
        // We need to get the latest revision, in the case where the published
        // revision is not the latest revision. For example if the programme
        // is published but we have a in-cycle draft forward revision.
        $revision_id = $node_storage->getLatestTranslationAffectedRevisionId($node->id(), $node->language()->getId());
        $latest_revision = $node_storage->loadRevision($revision_id);

        if ($this->configFactory->get('content_moderation_bulk_state_change.settings')->get('new_revision')) {
          /** @var \Drupal\psw_programme\Entity\ProgrammeInterface */
          $revision = $node_storage->createRevision($latest_revision, $latest_revision->isDefaultRevision());
          $revision->setRevisionCreationTime($this->time->getRequestTime());
        }
        else {
          $revision = $latest_revision;
          $revision->setNewRevision(FALSE);
        }

        $revision->setRevisionUserId($this->currentUser()->id());
        $this->setValues($revision, $form_state);
        }
      }

    $this->messenger()->addMessage($this->t('Entities have been successfully transitioned.'), MessengerInterface::TYPE_STATUS);
    $form_state->setRedirect('view.content.page_1');
  }

  /**
   * Check user has permission to edit all the entities.
   */
  private function userHasEditPermissions($entities): bool {
    foreach ($entities as $entity) {
      if (!$entity->access('update', $this->currentUser())) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  private function setValues(NodeInterface $revision, FormStateInterface $form_state) {
    $current_stage = $revision->get('moderation_state')->value;
    $to_form_state = $form_state->getValue('transition_to');

    $revision->set('moderation_state', $to_form_state);

    $to_state_label = $this->getWorkflowStageLabel($revision, $to_form_state);
    $from_state_label = $this->getWorkflowStageLabel($revision, $current_stage);

    $revision->setRevisionLogMessage($this->t('Moderation state changed from @from_state to @to_state.',
      ['@from_state' => $from_state_label, '@to_state' => $to_state_label]));

    $revision->save();
  }


  /**
   * Check entities all belong to the same workflow.
   */
  private function entitiesAllSameWorkflow($entities): bool {
    $moderation_workflow_ids = [];
    foreach ($entities as $entity) {
      $moderation_workflow_ids[] = $this->moderationInformation->getWorkflowForEntity($entity)->id();
    }
    return count(array_unique($moderation_workflow_ids)) === 1;
  }

  /**
   * Check entities all belong to the same workflow state.
   */
  private function entitiesAllSameWorkflowStage($entities): bool {
    $moderation_workflow_states = [];
    foreach ($entities as $entity) {
      $moderation_workflow_states[] = $entity->get('moderation_state')->value ?? '';
    }
    return count(array_unique($moderation_workflow_states)) === 1;
  }

  /**
   * Get the states a given moderation state can transition to.
   */
  private function getAvailableToStates($entity): array {
    $available_to_states = [];
    $workflow = $this->moderationInformation->getWorkflowForEntity($entity);
    $type_plugin = $workflow->getTypePlugin();
    $state_id = $entity->get('moderation_state')->value;
    $available_transitions = $type_plugin->getTransitionsForState($state_id, TransitionInterface::DIRECTION_FROM);

    foreach ($available_transitions as $available_transition) {
      if ($available_transition->to()->id() !== $state_id) {
        $available_to_states[$available_transition->to()->id()] = $available_transition->to()->label();
      }
    }

    return $available_to_states;
  }

  /**
   * Get user friendly label for workflow state.
   */
  private function getWorkflowStageLabel($entity, $state_id): string {
    $workflow_states = $this->moderationInformation->getWorkflowForEntity($entity)->getTypePlugin()->getStates();
    return $workflow_states[$state_id]->label();
  }

  /**
   * Redirect user back to listing.
   */
  public function redirectToListing(): ?RedirectResponse {
    $destination = $this->requestStack->getCurrentRequest()->get('destination');

    if ($destination) {
      $url = Url::fromUserInput('/' . ltrim($destination, '/'));
      return $this->redirect($url->getRouteName());
    }
    return NULL;
  }

}
